# -*- coding: utf-8 -*-
"""
pyHRV Version
-------------

Version tracker.

:copyright: (c) 2018 by Pedro Gomes (HAW Hamburg)
:license: BSD 3-clause, see LICENSE for more details.
"""

VERSION = (0, 4, 0)
__version__ = '.'.join(map(str, VERSION))
